package sun.rmi.server;

import java.io.IOException;
import java.rmi.server.RMISocketFactory;
import java.net.ServerSocket;
import java.rmi.server.RMIServerSocketFactory;
import java.rmi.activation.UnknownObjectException;
import java.rmi.NoSuchObjectException;
import java.rmi.activation.Activatable;
import java.rmi.activation.ActivationException;
import java.security.PrivilegedActionException;
import java.security.AccessController;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Constructor;
import java.security.PrivilegedExceptionAction;
import java.rmi.server.RMIClassLoader;
import sun.rmi.registry.RegistryImpl;
import java.rmi.Remote;
import java.rmi.activation.ActivationDesc;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.ArrayList;
import java.rmi.MarshalledObject;
import java.util.List;
import java.rmi.activation.ActivationGroupID;
import java.rmi.activation.ActivationID;
import java.util.Hashtable;
import java.rmi.activation.ActivationGroup;

public class ActivationGroupImpl extends ActivationGroup
{
    private static final long serialVersionUID = 5758693559430427303L;
    private final Hashtable<ActivationID, ActivationGroupImpl.ActiveEntry> active;
    private boolean groupInactive;
    private final ActivationGroupID groupID;
    private final List<ActivationID> lockedIDs;
    
    public ActivationGroupImpl(ActivationGroupID groupID, MarshalledObject<?> marshalledObject) throws RemoteException {
        super(groupID);
        this.active = new Hashtable();
        this.groupInactive = false;
        this.lockedIDs = new ArrayList();
        this.groupID = groupID;
        unexportObject(this, true);
        UnicastRemoteObject.exportObject(this, 0, null, new ActivationGroupImpl.ServerSocketFactoryImpl());
        if (System.getSecurityManager() == null) {
            try {
                System.setSecurityManager(new SecurityManager());
            }
            catch (Exception cause) {
                throw new RemoteException("unable to set security manager", cause);
            }
        }
    }
    
    private void acquireLock(ActivationID activationID) {
        while (true) {
            ActivationID activationID2;
            synchronized (this.lockedIDs) {
                int index = this.lockedIDs.indexOf(activationID);
                if (index < 0) {
                    this.lockedIDs.add(activationID);
                    return;
                }
                activationID2 = (ActivationID)this.lockedIDs.get(index);
            }
            synchronized (activationID2) {
                synchronized (this.lockedIDs) {
                    int index2 = this.lockedIDs.indexOf(activationID2);
                    if (index2 < 0) {
                        continue;
                    }
                    if ((ActivationID)this.lockedIDs.get(index2) != activationID2) {
                        continue;
                    }
                }
                try {
                    activationID2.wait();
                }
                catch (InterruptedException ex) {}
            }
        }
    }
    
    private void releaseLock(ActivationID activationID) {
        synchronized (this.lockedIDs) {
            activationID = (ActivationID)this.lockedIDs.remove(this.lockedIDs.indexOf(activationID));
        }
        synchronized (activationID) {
            activationID.notifyAll();
        }
    }
    
    public MarshalledObject<? extends Remote> newInstance(final ActivationID activationID, final ActivationDesc activationDesc) throws ActivationException, RemoteException {
        RegistryImpl.checkAccess("ActivationInstantiator.newInstance");
        if (!this.groupID.equals(activationDesc.getGroupID())) {
            throw new ActivationException("newInstance in wrong group");
        }
        try {
            this.acquireLock(activationID);
            synchronized (this) {
                if (this.groupInactive) {
                    throw new InactiveGroupException("group is inactive");
                }
            }
            ActivationGroupImpl.ActiveEntry activeEntry = (ActivationGroupImpl.ActiveEntry)this.active.get(activationID);
            if (activeEntry != null) {
                return activeEntry.mobj;
            }
            final Class<Object> subclass = RMIClassLoader.loadClass(activationDesc.getLocation(), activationDesc.getClassName()).asSubclass(Remote.class);
            Remote remote = null;
            final Thread currentThread = Thread.currentThread();
            final ClassLoader contextClassLoader = currentThread.getContextClassLoader();
            ClassLoader classLoader = subclass.getClassLoader();
            contextClassLoader = (covers(classLoader, contextClassLoader) ? classLoader : contextClassLoader);
            try {
                remote = (Remote)AccessController.doPrivileged(new PrivilegedExceptionAction<Remote>() {
                    @Override
                    public Remote run() throws InstantiationException, NoSuchMethodException, IllegalAccessException, InvocationTargetException {
                        Constructor<Object> declaredConstructor = subclass.getDeclaredConstructor(new Class[] { ActivationID.class, MarshalledObject.class });
                        declaredConstructor.setAccessible(true);
                        try {
                            currentThread.setContextClassLoader(contextClassLoader);
                            return (Remote)declaredConstructor.newInstance(new Object[] { activationID, activationDesc.getData() });
                        }
                        finally {
                            currentThread.setContextClassLoader(contextClassLoader);
                        }
                    }
                });
            }
            catch (PrivilegedActionException ex) {
                Exception exception = ex.getException();
                if ((exception instanceof InstantiationException)) {
                    throw (InstantiationException)exception;
                }
                if ((exception instanceof NoSuchMethodException)) {
                    throw (NoSuchMethodException)exception;
                }
                if ((exception instanceof IllegalAccessException)) {
                    throw (IllegalAccessException)exception;
                }
                if ((exception instanceof InvocationTargetException)) {
                    throw (InvocationTargetException)exception;
                }
                if ((exception instanceof RuntimeException)) {
                    throw (RuntimeException)exception;
                }
                if ((exception instanceof Error)) {
                    throw (Error)exception;
                }
            }
            ActivationGroupImpl.ActiveEntry value = new ActiveEntry(remote);
            this.active.put(activationID, value);
            return value.mobj;
        }
        catch (NoSuchMethodException | NoSuchMethodError noSuchMethodException | NoSuchMethodError) {
            throw new ActivationException("Activatable object must provide an activation constructor", noSuchMethodException | NoSuchMethodError);
        }
        catch (InvocationTargetException ex2) {
            throw new ActivationException("exception in object constructor", ex2.getTargetException());
        }
        catch (Exception ex3) {
            throw new ActivationException("unable to activate object", ex3);
        }
        finally {
            this.releaseLock(activationID);
            this.checkInactiveGroup();
        }
    }
    
    public boolean inactiveObject(ActivationID activationID) throws ActivationException, UnknownObjectException, RemoteException {
        try {
            this.acquireLock(activationID);
            synchronized (this) {
                if (this.groupInactive) {
                    throw new ActivationException("group is inactive");
                }
            }
            ActivationGroupImpl.ActiveEntry activeEntry = (ActivationGroupImpl.ActiveEntry)this.active.get(activationID);
            if (activeEntry == null) {
                throw new UnknownObjectException("object not active");
            }
            try {
                if (!Activatable.unexportObject(activeEntry.impl, false)) {
                    return false;
                }
            }
            catch (NoSuchObjectException ex) {}
            try {
                super.inactiveObject(activationID);
            }
            catch (UnknownObjectException ex2) {}
            this.active.remove(activationID);
        }
        finally {
            this.releaseLock(activationID);
            this.checkInactiveGroup();
        }
        return true;
    }
    
    private void checkInactiveGroup() {
        boolean n = false;
        synchronized (this) {
            if (this.active.size() == 0 && this.lockedIDs.size() == 0 && !this.groupInactive) {
                this.groupInactive = true;
                n = true;
            }
        }
        if (!(n == 0)) {
            try {
                super.inactiveGroup();
            }
            catch (Exception ex) {}
            try {
                UnicastRemoteObject.unexportObject(this, true);
            }
            catch (NoSuchObjectException ex2) {}
        }
    }
    
    public void activeObject(ActivationID activationID, Remote remote) throws ActivationException, UnknownObjectException, RemoteException {
        try {
            this.acquireLock(activationID);
            synchronized (this) {
                if (this.groupInactive) {
                    throw new ActivationException("group is inactive");
                }
            }
            if (!this.active.contains(activationID)) {
                ActivationGroupImpl.ActiveEntry value = new ActiveEntry(remote);
                this.active.put(activationID, value);
                try {
                    super.activeObject(activationID, value.mobj);
                }
                catch (RemoteException ex) {}
            }
        }
        finally {
            this.releaseLock(activationID);
            this.checkInactiveGroup();
        }
    }
    
    private static boolean covers(ClassLoader parent, ClassLoader classLoader) {
        if (classLoader == null) {
            return true;
        }
        if (parent == null) {
            return false;
        }
        while (parent != classLoader) {
            parent = parent.getParent();
            if (parent != null) {
                continue;
            }
            return false;
        }
        return true;
    }
    
    private static class ServerSocketFactoryImpl implements RMIServerSocketFactory
    {
        @Override
        public ServerSocket createServerSocket(int n) throws IOException {
            RMISocketFactory rmiSocketFactory = RMISocketFactory.getSocketFactory();
            if (rmiSocketFactory == null) {
                rmiSocketFactory = RMISocketFactory.getDefaultSocketFactory();
            }
            return rmiSocketFactory.createServerSocket(n);
        }
    }
    
    private static class ActiveEntry
    {
        Remote impl;
        MarshalledObject<Remote> mobj;
        
        ActiveEntry(Remote remote) throws ActivationException {
            this.impl = remote;
            try {
                this.mobj = new MarshalledObject(remote);
            }
            catch (IOException ex) {
                throw new ActivationException("failed to marshal remote object", ex);
            }
        }
    }
}
